/* -*- Mode: C++; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 *     Copyright 2011-2020 Couchbase, Inc.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

#include "config.h"
#include "internal.h"
#include <gtest/gtest.h>
#define LIBCOUCHBASE_INTERNAL 1
#include <libcouchbase/couchbase.h>

class MiscTests : public ::testing::Test
{
};

TEST_F(MiscTests, testGetTmpdir)
{
    const char *tmpdir = lcb_get_tmpdir();
    ASSERT_FALSE(tmpdir == NULL);
    ASSERT_STRNE("", tmpdir);
}
TEST_F(MiscTests, testVersionG)
{
    ASSERT_GT(lcb_version_g, (lcb_U32)0);
}
